function [labeledImage,nCells,originalImage] = labelCells(fileName,doPlot)
% function LABELCELLS - counts and labels cells in a single image
%
%   L = labelCells(FNAME) will read in an image with the filename FNAME and
%   count cells. The basic algorithm uses thresholding and labeling
%   connected regions with bwlabel(), but there are several preprocessing
%   steps involving morphological operations, e.g. imdilate
%
%   L = labelCells(FNAME,DO_PLOT) will display figures showing each
%   processing step if DO_PLOT = 1. The default value is 0
%
%   [L,NCELLS] = labelCells(...) will output the number of cells in
%   the labeled image as a scalar NCELLs. NCELLS = max(L(:))
%
%   [L,NCELLS,ORIG_IM] = labelCells(...) will also output the original
%   image in ORIG_IM. This image has been normalized between 0 and 1.

%% Handle input
if nargin<2
    doPlot = false;
end
   

%% Load and show the image
%Open the image
originalImage = imread(fileName);

%Convert to double and rescale (normalize) the image
originalImage = rescale(double(originalImage));

%Plot original image for comparison
if doPlot
    figure, imshow(originalImage)
    title([fileName ': original'],'Interpreter','none')
end

    
%% Initial opening to make cells uniform inside
% This obscures things like nuclei in the cells but preserves edges
% between cells

openedImage = imopen(originalImage,strel('disk',13));
if doPlot
    figure, imshow(openedImage)
    title([fileName ': opened'],'Interpreter','none')
end
    

%% Emphasize cell boundaries
% This code is taken directly from: 
% https://www.mathworks.com/help/vision/examples/cell-counting.html
% It exaggerates the borders between cells using two dilation steps and
% some arithmetic. It works well as long as cells are fairly uniform
% inside, which is why we have the intial imopen step.

SE = strel('disk',8);
y1 = 2*openedImage - imdilate(openedImage, SE);
y1(y1<0) = 0;
y1(y1>1) = 1;
dilatedImage = imdilate(y1, SE) - y1;


if doPlot
    figure, imshow(dilatedImage)
    title([fileName ': exaggerated boundaries'],'Interpreter','none')
end


%% Convert to binary (threshold)

% The 0.7 factor is also from the same matlab link above. A lower threshold 
% helps to even further exaggerate the boundaries between cells.
threshold = 0.7*graythresh(dilatedImage);

% We use <= because we want the cell bodies to be white and the borders to
% be black. Functions like im2bw will use >=
bwImage = dilatedImage <= threshold;

if doPlot
    figure, imshow(bwImage)
    title([fileName ': thresholded'],'Interpreter','none')
end


%% Label connected regions

% We do an intial labeling of all connected regions in the bwImage
[labeledImage,nCells] = bwlabel(bwImage);
  
if doPlot
   figure, imagesc(labeledImage)
   title([fileName ': ' num2str(nCells) ' cells labeled'],'Interpreter','none')
end


%% Remove outliers: big and small areas
cellProperties = regionprops(labeledImage,'Area');
cellAreas = [cellProperties.Area];

% Remove small areas and big areas. These numbers are unique to this data
% set. Other images will need different criteria. We use the find function
% here so smallIndexes can be the loop variables (wouldn't work for a
% logical array)
smallIndexes = find(cellAreas < 500 | cellAreas > 3000);

% Make a copy of labeledImage because the correction in the following loop
% makes changes to labeledImage
labeledCopy = labeledImage;

% Loop through smallIndexes, making corrections to labeledImage
for iSmall = smallIndexes
    
    % Find pixels in image that correspond to false positive and set them
    % to zero. We always compare with labeledCopy since we are editing
    % labeledIMage
    isSmall = labeledCopy == iSmall;
    labeledImage(isSmall) = 0;
    
    % To get a correct number of cells, we now need to subtract all the
    % labels higher than iSmall by 1
    needsFix = labeledCopy > iSmall;
    labeledImage(needsFix) = labeledImage(needsFix) - 1;
end

% Redo the labeled plot with correct number
nCells = max(labeledImage(:));
if doPlot
   figure, imagesc(labeledImage)
   title([fileName ': ' num2str(nCells) ' cells labeled after correction'], ...
       'Interpreter','none')
end


%% Final plot. Overlay output on original image
if doPlot
    overlaidImage = labeloverlay(originalImage,labeledImage);
    figure, imshow(overlaidImage)
    title([fileName ': overlay with ' num2str(nCells) ' cells'],'Interpreter','none')
end
